#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import csv
import os
import tempfile
from collections.abc import Sequence

import numpy as np
import pandas as pd
import scipy.io as sio
from sos.utils import env


def homogeneous_type(seq):
    iseq = iter(seq)
    first_type = type(next(iseq))
    if first_type in (int, float):
        return True if all(isinstance(x, (int, float)) for x in iseq) else False
    return True if all(isinstance(x, first_type) for x in iseq) else False


Matlab_init_statements = rf'''
path(path, {os.path.split(__file__)[0]!r})
'''


class sos_MATLAB:
    supported_kernels = {'MATLAB': ['imatlab', 'matlab'], 'Octave': ['octave']}
    background_color = {'MATLAB': '#8ee7f1', 'Octave': '#dff8fb'}
    options = {}
    cd_command = 'cd {dir}'

    def __init__(self, sos_kernel, kernel_name='matlab'):
        self.sos_kernel = sos_kernel
        self.kernel_name = kernel_name
        self.init_statements = Matlab_init_statements
        if self.kernel_name == 'octave':
            self.init_statements += 'pkg load dataframe\n'

    def _Matlab_repr(self, obj):
        #  Converting a Python object to a Matlab expression that will be executed
        #  by the Matlab kernel.
        if isinstance(obj, bool):
            return 'true' if obj else 'false'
        if isinstance(obj, (int, float, str, complex)):
            return repr(obj)
        if isinstance(obj, Sequence):
            if len(obj) == 0:
                return '[]'

            # if the data is of homogeneous type, let us use []

            if homogeneous_type(obj):
                return '[' + ';'.join(self._Matlab_repr(x) for x in obj) + ']'
            return '{' + ';'.join(self._Matlab_repr(x) for x in obj) + '}'
        if obj is None:
            return 'NaN'
        if isinstance(obj, dict):
            dic = tempfile.tempdir
            sio.savemat(os.path.join(dic, 'dict2mtlb.mat'), {'obj': obj})
            return 'getfield(load(fullfile(' + '\'' + dic + '\'' + ',' \
                + '\'dict2mtlb.mat\')), \'obj\')'

        if isinstance(obj, set):
            return '{' + ','.join(self._Matlab_repr(x) for x in obj) + '}'
        if isinstance(obj, (
                np.intc,
                np.intp,
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
                np.float16,
                np.float32,
                np.float64,
        )):
            return repr(obj)

        if isinstance(obj, np.matrixlib.defmatrix.matrix):
            dic = tempfile.tempdir
            sio.savemat(os.path.join(dic, 'mat2mtlb.mat'), {'obj': obj})
            return 'cell2mat(struct2cell(load(fullfile(' + '\'' + dic + '\'' + ',' \
                + '\'mat2mtlb.mat\'))))'
        if isinstance(obj, np.ndarray):
            dic = tempfile.tempdir
            sio.savemat(os.path.join(dic, 'ary2mtlb.mat'), {'obj': obj})
            return 'sos_load_obj(fullfile(' + '\'' + dic + '\'' + ',' \
                + '\'ary2mtlb.mat\'))'
        if isinstance(obj, pd.DataFrame):
            if self.kernel_name == 'octave':
                dic = tempfile.tempdir
                obj.to_csv(os.path.join(dic, 'df2oct.csv'), index=False, quoting=csv.QUOTE_NONNUMERIC, quotechar="'")
                return 'dataframe(' + '\'' + dic + '/' + 'df2oct.csv\')'
            dic = tempfile.tempdir
            obj.to_csv(os.path.join(dic, 'df2mtlb.csv'), index=False, quoting=csv.QUOTE_NONNUMERIC, quotechar="'")
            return 'readtable(' + '\'' + dic + '/' + 'df2mtlb.csv\')'

    async def get_vars(self, names, as_var=None):
        for name in names:
            # add 'm' to any variable beginning with '_'
            if as_var is not None:
                newname = as_var
            elif name.startswith('_'):
                self.sos_kernel.warn(f'Variable {name} is passed from SoS to kernel {self.kernel_name} as {"m" + name}')
                newname = 'm' + name
            else:
                newname = name
            matlab_repr = self._Matlab_repr(env.sos_dict[name])
            env.log_to_file('KERNEL', f'Executing \n{matlab_repr}')
            await self.sos_kernel.run_cell(
                f'{newname} = {matlab_repr}',
                True,
                False,
                on_error=f'Failed to get variable {name} of type {env.sos_dict[name].__class__.__name__} to Matlab')

    def put_vars(self, items, to_kernel=None, as_var=None):
        if not items:
            return {}

        result = {}
        for item in items:
            py_repr = f'display(sos_py_repr({item}))'
            #9 MATLAB can use multiple messages for standard output,
            # so we need to concatenate these outputs.
            expr = ''
            for _, msg in self.sos_kernel.get_response(py_repr, ('stream',), name=('stdout',)):
                expr += msg['text']

            cwd = os.getcwd()
            try:
                if 'loadmat' in expr:
                    # imported to be used by eval
                    from scipy.io import loadmat
                # evaluate as raw string to correctly handle \\ etc
                result[as_var if as_var else item] = eval(expr)
            except Exception as e:
                self.sos_kernel.warn(f'Failed to evaluate {expr!r}: {e}')
            finally:
                os.chdir(cwd)
        return result

    def sessioninfo(self):
        return self.sos_kernel.get_response(r'ver', ('stream',), name=('stdout',))[0][1]['text']



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

from sos_notebook.test_utils import NotebookTest
import random

import shutil
import pytest

@pytest.mark.skipif(not shutil.which('octave'), reason='Octave is not available')
class TestOctaveDataExchange(NotebookTest):

    def _var_name(self):
        if not hasattr(self, '_var_idx'):
            self._var_idx = 0
        self._var_idx += 1
        return f'var{self._var_idx}'

    def get_from_SoS(self, notebook, sos_expr):
        var_name = self._var_name()
        notebook.call(f'{var_name} = {sos_expr}', kernel='SoS')
        return notebook.check_output(
            f'''\
            %get {var_name}
            disp({var_name})
            ''',
            kernel='Octave')

    def put_to_SoS(self, notebook, py2_expr):
        var_name = self._var_name()
        notebook.call(
            f'''\
            %put {var_name}
            {var_name} = {py2_expr}
            ''',
            kernel='Octave')
        return notebook.check_output(f'print(repr({var_name}))', kernel='SoS')

    def test_get_none(self, notebook):
        assert 'NaN' == self.get_from_SoS(notebook, 'None')

    def test_put_NaN(self, notebook):
        assert 'None' == self.put_to_SoS(notebook, 'NaN')

    def test_get_int(self, notebook):
        assert 123 == int(self.get_from_SoS(notebook, '123'))
        assert '1.2346e+12' == self.get_from_SoS(notebook, '1234567891234')

    def test_put_int(self, notebook):
        assert 123 == int(self.put_to_SoS(notebook, '123'))
        assert 1234567891234 == int(self.put_to_SoS(notebook, '1234567891234'))
        # rounding error occurs
        assert 123456789123456784 == int(
            self.put_to_SoS(notebook, '123456789123456789'))

    def test_get_double(self, notebook):
        val = str(random.random())
        notebook.call('format long', kernel='Octave')
        assert abs(float(val) - float(self.get_from_SoS(notebook, val))) < 1e-10

    def test_put_double(self, notebook):
        val = str(random.random())
        notebook.call('format long', kernel='Octave')
        assert abs(float(val) - float(self.put_to_SoS(notebook, val))) < 1e-10

    def test_get_logic(self, notebook):
        assert '1' == self.get_from_SoS(notebook, 'True')
        assert '0' == self.get_from_SoS(notebook, 'False')

    def test_put_logic(self, notebook):
        assert 'True' == self.put_to_SoS(notebook, 'true')
        assert 'False' == self.put_to_SoS(notebook, 'false')

    def test_get_num_array(self, notebook):
        assert '1' == self.get_from_SoS(notebook, '[1]')
        assert '1\n2' == self.get_from_SoS(notebook, '[1, 2]').replace(' ', '')
        #
        notebook.call('format short', kernel='Octave')
        assert '1.2300' == self.get_from_SoS(notebook, '[1.23]')
        assert '1.4000\n2.0000' == self.get_from_SoS(notebook,
                                                     '[1.4, 2]').replace(
                                                         ' ', '')

    def test_get_numpy_array(self, notebook):
        notebook.call('import numpy as np', kernel='SoS')
        assert '1  2  3' == self.get_from_SoS(notebook, 'np.array([1, 2, 3])')
        notebook.call('format short', kernel='Octave')
        assert '11.0000    2.1000   32.0000' == self.get_from_SoS(
            notebook, 'np.array([11, 2.1, 32])')

    def test_put_num_array(self, notebook):
        assert '1' == self.put_to_SoS(notebook, '[1]')
        assert 'array([1, 2])' == self.put_to_SoS(notebook, '[1, 2]')
        #
        assert '1.23' == self.put_to_SoS(notebook, '[1.23]')
        assert 'array([1.4, 2. ])' == self.put_to_SoS(notebook, '[1.4, 2]')

    def test_get_logic_array(self, notebook):
        assert '1' == self.get_from_SoS(notebook, '[True]')
        assert '1\n0\n1' == self.get_from_SoS(notebook,
                                              '[True, False, True]').replace(
                                                  ' ', '')

    def test_put_logic_array(self, notebook):
        # Note that single element numeric array is treated as single value
        assert 'True' == self.put_to_SoS(notebook, '[true]')
        assert '[True, False, True]' == self.put_to_SoS(notebook,
                                                        '[true, false, true]')

    def test_get_str(self, notebook):
        assert "ab c d" == self.get_from_SoS(notebook, "'ab c d'")
        assert "ab\\td" == self.get_from_SoS(notebook, r"'ab\td'")

    def test_put_str(self, notebook):
        assert "'ab c d'" == self.put_to_SoS(notebook, "'ab c d'")
        assert r"'ab\\td'" == self.put_to_SoS(notebook, r"'ab\td'")

    def test_get_str_array(self, notebook):
        output = self.get_from_SoS(notebook, "['a1', 'a2', 'a3']")
        assert 'a1' in output and 'a2' in output and 'a3' in output

    def test_put_str_array(self, notebook):
        assert "['a1', 'a2', 'a3cv']" == self.put_to_SoS(
            notebook, "['a1'; 'a2'; 'a3cv']")

    def test_get_mixed_list(self, notebook):
        output = self.get_from_SoS(notebook, '[2.4, True, "asd"]')
        assert '2.4' in output and '1' in output and 'asd' in output

    def test_get_dict(self, notebook):
        output = self.get_from_SoS(notebook, "dict(a=1, b='2')")
        assert 'scalar structure' in output and 'a = 1' in output and 'b = 2' in output

    def test_get_set(self, notebook):
        output = self.get_from_SoS(notebook, "{1.5, 'abc'}")
        assert '1.5' in output or abc in output

    def test_get_complex(self, notebook):
        assert "1.0000 + 2.2000i" == self.get_from_SoS(notebook,
                                                       "complex(1, 2.2)")

    def test_put_complex(self, notebook):
        assert "(2+3j)" == self.put_to_SoS(notebook, "2+3i")

    def test_get_recursive(self, notebook):
        output = self.get_from_SoS(notebook,
                                   "{'a': 1, 'b': {'c': 3, 'd': 'whatever'}}")
        assert 'scalar structure' in output and 'a = 1' in output and 'c = 3' in output and 'd = whatever' in output

    def test_get_matrix(self, notebook):
        notebook.call('import numpy as np', kernel='SoS')
        output = self.get_from_SoS(notebook, 'np.matrix([[11,22],[33,44]])')
        assert all(x in output for x in ('11', '22', '33', '44'))

    def test_put_matrix(self, notebook):
        output = self.put_to_SoS(notebook, '[1:3; 2:4]')
        assert 'matrix' in output and '[1., 2., 3.]' in output and '[2., 3., 4.]]' in output

    def test_get_ndarray(self, notebook):
        notebook.call(
            '''\
            %put var_3d --to Octave
            from numpy import zeros
            var_3d = zeros([2, 3, 4])
        ''',
            kernel='SoS')
        assert ['2', '3', '4'] == notebook.check_output(
            'disp(size(var_3d))', kernel='Octave').split()

    def test_put_ndarray(self, notebook):
        notebook.call(
            '''\
            %put octave_var_3d
            octave_var_3d = zeros([2, 3, 4])
        ''',
            kernel='Octave')
        assert '(2, 3, 4)' == notebook.check_output(
            'octave_var_3d.shape', kernel='SoS')

    def test_get_dataframe(self, notebook):
        notebook.call(
            '''
            %put df --to Octave
            import pandas as pd
            import numpy as np
            arr = np.random.randn(100)
            df = pd.DataFrame({'column_{0}'.format(i): arr for i in range(10)})
            ''',
            kernel='SoS')
        output = notebook.check_output('df', kernel='Octave')
        assert 'dataframe' in output and '100 rows' in output
        #
        #  non-numeric dataframe
        notebook.call(
            '''
            %put df --to Octave
            import pandas as pd

            df = pd.DataFrame({'name': ['Leonardo', 'Donatello', 'Michelangelo', 'Raphael'],
                   'num': [1, 2, 3, 4],
                   'weapon': ['ninjatos', 'bo', 'nunchaku', 'sai']})
            ''',
            kernel='SoS')
        output = notebook.check_output('df', kernel='Octave')
        assert 'dataframe' in output and '4 rows' in output and '3 columns' in output and 'Michelangelo' in output



#!/usr/bin/env javascript
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

from setuptools import find_packages, setup

# obtain version of SoS
with open('src/sos_matlab/_version.py') as version:
    for line in version:
        if line.startswith('__version__'):
            __version__ = eval(line.split('=')[1])
            break

setup(name = "sos-matlab",
    version = __version__,
    description = 'SoS Notebook extension for Matlab and Octave',
    author = 'Bo Peng',
    url = 'https://github.com/vatlab/SOS',
    author_email = 'bpeng@mdanderson.org',
    maintainer = 'Bo Peng',
    maintainer_email = 'bpeng@mdanderson.org',
    license = '3-clause BSD',
    include_package_data = True,
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        ],
    packages = find_packages('src'),
    package_dir = {'': 'src'},
    install_requires=[
          'sos>=0.18.0',
          'sos-notebook>=0.24.0',
          'scipy',
      ],
    entry_points= '''
[sos_languages]
MATLAB = sos_matlab.kernel:sos_MATLAB
Octave = sos_matlab.kernel:sos_MATLAB
[sos_converters]
m-ipynb = sos_matlab.converter:OctaveToNotebookConverter
'''
)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import tempfile
import pytest
from sos_notebook.test_utils import NotebookTest

import shutil
import pytest


@pytest.mark.skipif(
    not shutil.which('matlab'), reason='MATLAB is not available')
class TestMATLABInterface(NotebookTest):

    #
    # Python 2
    #
    def test_prompt_color(self, notebook):
        '''test color of input and output prompt'''
        idx = notebook.call(
            '''\
            disp('this is MATLAB')
            ''',
            kernel="MATLAB")
        assert [142, 231, 241] == notebook.get_input_backgroundColor(idx)
        assert [142, 231, 241] == notebook.get_output_backgroundColor(idx)

    def test_cd(self, notebook):
        '''Support for change of directory with magic %cd'''
        output1 = notebook.check_output(
            '''\
            disp(pwd)
            ''', kernel="MATLAB")
        notebook.call('%cd ..', kernel="SoS")
        output2 = notebook.check_output(
            '''\
            disp(pwd)
            ''', kernel="MATLAB")
        assert len(output1) > len(output2)
        assert output1.startswith(output2)
        #
        # cd to a specific directory
        tmpdir = os.path.join(tempfile.gettempdir(), 'somedir')
        os.makedirs(tmpdir, exist_ok=True)
        notebook.call(f'%cd {tmpdir}', kernel="SoS")
        output = notebook.check_output(
            '''\
            disp(pwd)
            ''', kernel="MATLAB")
        assert os.path.realpath(tmpdir) == os.path.realpath(output)

    # def test_preview(self, notebook):
    #     '''Test support for %preview'''
    #     output = notebook.check_output(
    #         '''\
    #         %preview -n var
    #         var = list(range(100))
    #         ''',
    #         kernel="MATLAB")
    #     # in a normal var output, 100 would be printed. The preview version would show
    #     # type and some of the items in the format of
    #     #   int [1:1000] 1 2 3 4 5 6 7 8 9 10 ...
    #     assert 'list' in output and '100' in output and '1' in output and '99' not in output

    #     #
    #     # return 'Unknown variable' for unknown variable
    #     assert 'Unknown variable' in notebook.check_output(
    #         '%preview -n unknown_var', kernel="MATLAB")
    #     #
    #     # return 'Unknown variable for expression
    #     assert 'Unknown variable' in notebook.check_output(
    #         '%preview -n var[1]', kernel="MATLAB")

    def test_sessioninfo(self, notebook):
        '''test support for %sessioninfo'''
        notebook.call("disp('this is MATLAB')", kernel="MATLAB")
        assert 'MATLAB' in notebook.check_output('%sessioninfo', kernel="SoS")
